#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import os
import threading

from . import consts
from . import logging as pc_logging
from . import project_config
from . import runtime_python_all
from . import project_factory_local as rfl
from . import project_factory_git as rfg
from . import project_factory_tar as rft
from .user_config import user_config
from .utils import *


# Context
class Context(project_config.Configuration):
    """Stores and caches all imported objects."""

    stats_packages: int
    stats_packages_instantiated: int
    stats_parts: int
    stats_parts_instantiated: int
    stats_assemblies: int
    stats_assemblies_instantiated: int
    stats_memory: int

    def __init__(self, root_path=None, search_root=True):
        """Initializes the context and imports the root project."""
        root_file = ""
        if root_path is None:
            # Find the top folder containing "partcad.yaml"
            root_path = "."
        else:
            if os.path.isfile(root_path):
                root_file = os.path.basename(root_path)
                root_path = os.path.dirname(root_path)
        initial_root_path = os.path.abspath(root_path)
        if search_root:
            while os.path.exists(os.path.join(root_path, "..", "partcad.yaml")):
                root_path = os.path.join(root_path, "..")
        self.root_path = os.path.abspath(root_path)
        if self.root_path == initial_root_path and root_file != "":
            self.root_path = os.path.join(self.root_path, root_file)
        self.current_project_path = "/" + os.path.relpath(
            initial_root_path,
            root_path,
        )
        if self.current_project_path == "/.":
            self.current_project_path = "/"
        super().__init__(consts.ROOT, self.root_path)

        # Protect the critical sections from access in different threads
        self.lock = threading.RLock()

        self.option_create_dirs = False
        self.runtimes_python = {}

        self.stats_packages = 0
        self.stats_packages_instantiated = 0
        self.stats_parts = 0
        self.stats_parts_instantiated = 0
        self.stats_assemblies = 0
        self.stats_assemblies_instantiated = 0
        self.stats_memory = 0

        # self.projects contains all projects known to this context
        self.projects = {}
        self._projects_being_loaded = {}

        with pc_logging.Process("InitCtx", self.config_dir):
            self.import_project(
                None,  # parent
                {
                    "name": consts.ROOT,
                    "type": "local",
                    "path": self.config_path,
                    "maybeEmpty": True,
                },
            )

    def stats_recalc(self, verbose=False):
        self.stats_memory = total_size(self, verbose)

    def get_current_project_path(self):
        return self.current_project_path

    def import_project(self, parent, project_import_config):
        if (
            not "name" in project_import_config
            or not "type" in project_import_config
        ):
            pc_logging.error(
                "Invalid project configuration found: %s"
                % project_import_config
            )
            return None

        name = project_import_config["name"]
        with pc_logging.Action("Import", name):
            if name in self._projects_being_loaded:
                pc_logging.error(
                    "Recursive project loading detected (%s), aborting." % name
                )
                return None
            self._projects_being_loaded[name] = True

            # Depending on the project type, use different factories
            if (
                not "type" in project_import_config
                or project_import_config["type"] == "local"
            ):
                rfl.ProjectFactoryLocal(self, parent, project_import_config)
            elif project_import_config["type"] == "git":
                with pc_logging.Action("Git", name):
                    rfg.ProjectFactoryGit(self, parent, project_import_config)
            elif project_import_config["type"] == "tar":
                with pc_logging.Action("Tar", name):
                    rft.ProjectFactoryTar(self, parent, project_import_config)
            else:
                pc_logging.error("Invalid project type found: %s." % name)
                del self._projects_being_loaded[name]
                return None

            # Check whether the factory was able to successfully add the project
            if not name in self.projects:
                pc_logging.error(
                    "Failed to create the project: %s" % project_import_config
                )
                del self._projects_being_loaded[name]
                return None

            self.stats_packages += 1
            self.stats_packages_instantiated += 1

            imported_project = self.projects[name]

            del self._projects_being_loaded[name]
            return imported_project

    def get_project_abs_path(self, rel_project_path: str):
        if rel_project_path.startswith("/"):
            return rel_project_path

        project_path = self.current_project_path

        if rel_project_path == ".":
            rel_project_path = ""
        if rel_project_path == "":
            return project_path

        if not project_path.endswith("/"):
            project_path += "/"
        return os.path.abspath(project_path + rel_project_path)

    def get_project(self, rel_project_path: str):
        project_path = self.get_project_abs_path(rel_project_path)

        with self.lock:
            # Strip the first '/' (absolute path always starts with a '/'``)
            project_path = project_path[1:]

            project = self.projects[consts.ROOT]

            if project_path == "":
                return project
            else:
                import_list = project_path.split("/")

            return self._get_project_recursive(project, import_list)

    def _get_project_recursive(self, project, import_list: list[str]):
        """Load the dependencies recursively"""
        if len(import_list) == 0:
            # Found what we are looking for
            return project

        # next_import is the next of the import we need to load now
        next_import = import_list[0]
        # import_list is reduced to contain only the items that will remain to
        # bt loaded after this import
        import_list = import_list[1:]

        # next_project will reference the project we are importing now
        next_project = None
        # next_project_path is the full path of the project we are importing now
        next_project_path = get_child_project_path(project.name, next_import)

        # see if the wanted project is already initialized
        if next_project_path in self.projects:
            return self._get_project_recursive(
                self.projects[next_project_path], import_list
            )

        # Check if there is a matching subfolder
        subfolders = [
            f.name for f in os.scandir(project.config_dir) if f.is_dir()
        ]
        if next_import in list(subfolders):
            if os.path.exists(
                os.path.join(
                    project.config_dir,
                    next_import,
                    consts.DEFAULT_PACKAGE_CONFIG,
                )
            ):
                pc_logging.debug(
                    "Importing a subfolder: %s..." % next_project_path
                )
                prj_conf = {
                    "name": next_project_path,
                    "type": "local",
                    "path": next_import,
                }
                next_project = self.import_project(project, prj_conf)
                if not next_project is None:
                    result = self._get_project_recursive(
                        next_project, import_list
                    )
                    return result
        else:
            # Otherwise, iterate all subfolders and check if any of them are packages
            if (
                "import" in project.config_obj
                and not project.config_obj["import"] is None
            ):
                for prj_name in project.config_obj["import"]:
                    pc_logging.debug("Checking the import: %s..." % prj_name)
                    if prj_name != next_import:
                        continue
                    pc_logging.debug("Importing: %s..." % next_project_path)
                    prj_conf = project.config_obj["import"][prj_name]
                    if "name" in prj_conf:
                        prj_conf["orig_name"] = prj_conf["name"]
                    prj_conf["name"] = next_project_path
                    next_project = self.import_project(project, prj_conf)
                    if not next_project is None:
                        result = self._get_project_recursive(
                            next_project, import_list
                        )
                        return result
                    break

        return next_project

    def import_all(self):
        self._import_all_recursive(self.projects[consts.ROOT])

    def _import_all_recursive(self, project):
        # First, iterate all explicitly mentioned "import"s.
        # Do it before iterating subdirectories, as it may kick off a long
        # background task.
        if (
            "import" in project.config_obj
            and not project.config_obj["import"] is None
        ):
            for prj_name in project.config_obj["import"]:
                next_project_path = get_child_project_path(
                    project.name, prj_name
                )

                pc_logging.debug("Importing: %s..." % next_project_path)
                prj_conf = project.config_obj["import"][prj_name]
                if "name" in prj_conf:
                    prj_conf["orig_name"] = prj_conf["name"]
                prj_conf["name"] = next_project_path
                next_project = self.import_project(project, prj_conf)
                self._import_all_recursive(next_project)

        # Second, iterate over all subfolder and check for packages
        subfolders = [
            f.name for f in os.scandir(project.config_dir) if f.is_dir()
        ]
        for subdir in list(subfolders):
            if os.path.exists(
                os.path.join(
                    project.config_dir,
                    subdir,
                    consts.DEFAULT_PACKAGE_CONFIG,
                )
            ):
                next_project_path = get_child_project_path(project.name, subdir)
                pc_logging.debug(
                    "Importing a subfolder: %s..." % next_project_path
                )
                prj_conf = {
                    "name": next_project_path,
                    "type": "local",
                    "path": subdir,
                }
                next_project = self.import_project(project, prj_conf)
                self._import_all_recursive(next_project)

    def get_all_packages(self):
        self.import_all()
        return self.get_packages()

    def get_packages(self):
        return map(
            lambda pkg: {"name": pkg.name, "desc": pkg.desc},
            filter(
                lambda x: len(x.parts) + len(x.assemblies) > 0,
                self.projects.values(),
            ),
        )

    def _get_part(self, part_spec, params=None):
        project_name, part_name = resolve_resource_path(
            self.current_project_path,
            part_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            pc_logging.error("Packages found: %s" % str(self.projects))
            return None
        pc_logging.debug("Retrieving %s from %s" % (part_name, project_name))
        return prj.get_part(part_name, params)

    def get_part(self, part_spec, params=None):
        return self._get_part(part_spec, params)

    def get_part_shape(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_wrapped())

    def get_part_cadquery(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_cadquery())

    def get_part_build123d(self, part_spec, params=None):
        return asyncio.run(self._get_part(part_spec, params).get_build123d())

    def _get_assembly(self, assembly_spec, params=None):
        project_name, assembly_name = resolve_resource_path(
            self.current_project_path,
            assembly_spec,
        )
        prj = self.get_project(project_name)
        if prj is None:
            pc_logging.error("Package %s not found" % project_name)
            return None
        pc_logging.debug(
            "Retrieving %s from %s" % (assembly_name, project_name)
        )
        return prj.get_assembly(assembly_name, params)

    def get_assembly(self, assembly_spec, params=None):
        return self._get_assembly(assembly_spec, params)

    def get_assembly_shape(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_wrapped()
        )

    def get_assembly_cadquery(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_cadquery()
        )

    def get_assembly_build123d(self, assembly_spec, params=None):
        return asyncio.run(
            self._get_assembly(assembly_spec, params).get_build123d()
        )

    def render(self, format=None, output_dir=None):
        path = self.get_current_project_path()
        pc_logging.debug("Rendering %s..." % path)
        prj = self.get_project(path)
        # prj = self.projects[path]
        prj.render(format=format, output_dir=output_dir)

    def get_python_runtime(self, version, python_runtime=None):
        if python_runtime is None:
            python_runtime = user_config.python_runtime
        runtime_name = python_runtime + "-" + version
        if not runtime_name in self.runtimes_python:
            self.runtimes_python[runtime_name] = runtime_python_all.create(
                self, version, python_runtime
            )
        return self.runtimes_python[runtime_name]

    def ensure_dirs(self, path):
        if not self.option_create_dirs:
            return
        os.makedirs(path)

    def ensure_dirs_for_file(self, filename):
        if not self.option_create_dirs:
            return
        path = os.path.dirname(filename)
        os.makedirs(path)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-01
#
# Licensed under Apache License, Version 2.0.
#

# This script contains code shared by all wrapper scripts.

import base64

# import fcntl # TODO(clairbee): replace it with whatever works on Windows if needed
import os
import pickle
import sys

from cq_serialize import register as register_cq_helper


def handle_input():
    if len(sys.argv) < 2:
        sys.stderr.write("Usage: %s <path>\n" % sys.argv[0])
        sys.exit(1)

    # Handle the input
    # - Comand line parameters
    path = sys.argv[1]
    if len(sys.argv) > 2:
        os.chdir(sys.argv[2])
    # - Content passed via stdin
    # #   - Make stdin blocking so that we can read until EOF
    # flag = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    # fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flag & ~os.O_NONBLOCK)
    #   - Read until EOF
    input_str = sys.stdin.read()
    #   - Unpack the content received via stdin
    request_bytes = base64.b64decode(input_str.encode())
    request = pickle.loads(request_bytes)
    return path, request


def handle_output(model):
    # Serialize the output
    register_cq_helper()
    picklestring = pickle.dumps(model)
    response = base64.b64encode(picklestring)
    return response.decode()



#
# OpenVMP, 2023-2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-01
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within the python sandbox environment (python runtime)
# to invoke `build1213d` scripts.

import os
import sys

from cadquery import cqgi

sys.path.append(os.path.dirname(__file__))
import wrapper_common


def process(path, request):
    build_parameters = {}
    if "build_parameters" in request:
        build_parameters = request["build_parameters"]

    script = open(path, "r").read()
    if "import partcad" in script:
        script = (
            "import logging\nlogging.basicConfig(level=60)\n"  # Disable PartCAD logging
            + script
        )
    script_object = cqgi.parse(script)
    result = script_object.build(build_parameters=build_parameters)

    if not result.success:
        sys.stderr.write("Exception: ")
        sys.stderr.write(str(result.exception))

    shape = None if not result.first_result else result.first_result.shape
    if hasattr(shape, "part"):
        shape = shape.part
    if hasattr(shape, "solid"):
        shape = shape.solid()
    if hasattr(shape, "wrapped"):
        shape = shape.wrapped

    return {
        "success": result.success,
        "exception": result.exception,
        "shape": shape,
    }


path, request = wrapper_common.handle_input()

# Call build123d through CQGI
model = process(path, request)

output = wrapper_common.handle_output(model)
print(output)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#


import os

from . import assembly as p
from . import assembly_factory as pf


class AssemblyFactoryFile(pf.AssemblyFactory):
    assembly: p.Assembly

    def __init__(self, ctx, project, assembly_config, extension=""):
        super().__init__(ctx, project, assembly_config)

        if "path" in assembly_config:
            self.path = assembly_config["path"]
        else:
            self.path = self.orig_name + extension

        if not os.path.isdir(project.config_dir):
            raise Exception(
                "ERROR: The project config directory must be a directory, found: '%s'"
                % project.config_dir
            )
        self.path = os.path.join(project.config_dir, self.path)
        if not os.path.exists(self.path):
            raise Exception("ERROR: The assembly path must exist")



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-23
#
# Licensed under Apache License, Version 2.0.
#

import hashlib
import inspect
import os
import requests
import tarfile

from . import project_factory as pf
from .user_config import user_config


class TarImportConfiguration:
    def __init__(self):
        self.import_config_url = self.config_obj.get("url")
        self.import_rel_path = self.config_obj.get("relPath")
        if "username" in self.config_obj and "password" in self.config_obj:
            self.auth_user = self.config_obj.get("username")
            self.auth_pass = self.config_obj.get("password")
        else:
            self.auth_user = None
            self.auth_pass = None


class ProjectFactoryTar(pf.ProjectFactory, TarImportConfiguration):
    def __init__(self, ctx, parent, config):
        pf.ProjectFactory.__init__(self, ctx, parent, config)
        TarImportConfiguration.__init__(self)

        # TODO(clairbee): Clone self.import_config_url to self.path
        self.path = self._extract(self.import_config_url)

        # Complement the config object here if necessary
        self._create(config)

        # TODO(clairbee): actually fill in the self.project object here

        self._save()

    def _extract(self, tarball_url, cache_dir=None):
        """
        Extracts a tarball to a local directory.

        Args:
          tarball_url: URL of the '.tar.gz' file to download.
          cache_dir: Directory to store the extracted files to (defaults to ".cache").

        Returns:
          Local path to the extracted content.
        """

        if cache_dir is None:
            cache_dir = os.path.join(user_config.internal_state_dir, "tar")

        # Generate a unique identifier for the file based on its URL.
        url_hash = hashlib.sha256(tarball_url.encode()).hexdigest()
        cache_path = os.path.join(cache_dir, url_hash)

        # Check if the tarball is already cached.
        if not os.path.exists(cache_path):
            # Download and extract
            try:
                os.makedirs(cache_path)

                auth = None
                if not (self.auth_user is None or self.auth_pass is None):
                    auth = (self.auth_user, self.auth_pass)
                with requests.get(
                    tarball_url, stream=True, auth=auth
                ) as rx, tarfile.open(fileobj=rx.raw, mode="r:gz") as tarobj:
                    args = inspect.getfullargspec(tarobj.extractall)

                    if "filter" in args.args:
                        if not self.import_rel_path is None:
                            filter = lambda member, _: (
                                member
                                if member.name.startswith(self.import_rel_path)
                                else None
                            )
                        else:
                            filter = lambda member, _: member

                        tarobj.extractall(cache_path, filter=filter)
                    else:
                        tarobj.extractall(cache_path)
            except Exception as e:
                raise RuntimeError(f"Failed to download the tarball: {e}")

        if not self.import_rel_path is None:
            cache_path = os.path.join(cache_path, self.import_rel_path)

        return cache_path



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os
from pathlib import Path
import shutil
import yaml

from . import logging as pc_logging


class UserConfig:
    @staticmethod
    def get_config_dir():
        return os.path.join(Path.home(), ".partcad")

    def __init__(self):
        self.config_obj = {}

        config_path = os.path.join(
            UserConfig.get_config_dir(),
            "config.yaml",
        )
        if os.path.exists(config_path):
            try:
                self.config_obj = yaml.safe_load(open(config_path, "r"))
            except Exception as e:
                pc_logging.error("ERROR: Failed to parse %s" % config_path)

        # option: pythonSandbox
        # description: sandboxing environment for invoking python scripts
        # values: [none | pypy | conda]
        # default: conda
        if not shutil.which("conda") is None:
            self.python_runtime = "conda"
        else:
            self.python_runtime = "none"

        if "pythonSandbox" in self.config_obj:
            self.python_runtime = self.config_obj["pythonSandbox"]

        # option: internalStateDir
        # description: folder to store all temporary files
        # values: <path>
        # default: '.partcad' folder in the home directory
        if "internalStateDir" in self.config_obj:
            self.internal_state_dir = self.config_obj["internalStateDir"]
        else:
            self.internal_state_dir = UserConfig.get_config_dir()

        # option: forceUpdate
        # description: update all repositories even if they are fresh
        # values: [True | False]
        # default: False
        if "forceUpdate" in self.config_obj:
            self.force_update = self.config_obj["forceUpdate"]
        else:
            self.force_update = False


user_config = UserConfig()



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import math
import typing

from . import shape
from . import sync_threads as pc_thread


class Part(shape.Shape):
    path: typing.Optional[str] = None

    def __init__(self, config: object = {}, shape=None):
        super().__init__(config)

        self.shape = shape
        self.lock = asyncio.Lock()

        self.desc = None
        if "desc" in config:
            self.desc = config["desc"]
        self.vendor = None
        if "vendor" in config:
            self.vendor = config["vendor"]
        self.sku = None
        if "sku" in config:
            self.sku = config["sku"]
        self.url = None
        if "url" in config:
            self.url = config["url"]
        if "count_per_sku" in config:
            self.count_per_sku = config["count_per_sku"]
        else:
            self.count_per_sku = 1
        self.count = 0

    async def get_shape(self):
        async with self.lock:
            if self.shape is None:
                self.shape = await pc_thread.run(self.instantiate, self)
            return self.shape

    def ref_inc(self):
        # TODO(clairbee): add a thread lock here
        self.count += 1

    def clone(self):
        cloned = Part(self.name, self.config, self.shape)
        cloned.count = self.count
        return cloned

    def _render_txt_real(self, file):
        file.write(self.name + ": " + self.count + "\n")

    def _render_markdown_real(self, file):
        name = self.name
        vendor = ""
        sku = ""
        if not self.desc is None:
            name = self.desc
        if not self.vendor is None:
            vendor = self.vendor
        if not self.sku is None:
            sku = self.sku
        if self.url is None:
            label = name
        else:
            label = "[" + name + "](" + self.url + ")"
            sku = "[" + sku + "](" + self.url + ")"
        count = str(math.ceil(self.count / self.count_per_sku))
        img_url = self._get_svg_url()

        file.write(
            "| "
            + label
            + " | "
            + count
            + " |"
            + vendor
            + " |"
            + sku
            + " |"
            + " !["
            + name
            + "]("
            + img_url
            + ")"
            + " |\n"
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.

import sys
from types import ModuleType, FunctionType
from gc import get_referents

from . import consts

# Custom objects know their class.
# Function objects seem to know way too much, including modules.
# Exclude modules as well.
BLACKLIST = type, ModuleType, FunctionType


def get_child_project_path(parent_path, child_name):
    if parent_path.endswith("/"):
        return parent_path + child_name
    else:
        return parent_path + "/" + child_name


def resolve_resource_path(current_project_name, pattern: str):
    if ":" in pattern:
        project_pattern, item_pattern = pattern.split(":")
        if project_pattern == "":
            project_pattern = current_project_name
    else:
        project_pattern = pattern
        item_pattern = "*"

    project_pattern = project_pattern.replace("...", "*")
    item_pattern = item_pattern.replace("...", "*")

    return project_pattern, item_pattern


def total_size(obj, verbose=False):
    """sum size of object & members."""
    if isinstance(obj, BLACKLIST):
        raise TypeError(
            "getsize() does not take argument of type: " + str(type(obj))
        )
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, BLACKLIST) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                s = sys.getsizeof(obj)
                if verbose:
                    print(s, type(obj), repr(obj), file=sys.stderr)
                size += s
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import copy
import os

# from pprint import pformat
import ruamel.yaml
import threading
import typing

from . import factory
from . import logging as pc_logging
from . import project_config
from . import part
from . import part_config
from . import part_factory_scad as pfscad
from . import part_factory_step as pfs
from . import part_factory_stl as pfstl
from . import part_factory_3mf as pf3
from . import part_factory_cadquery as pfc
from . import part_factory_build123d as pfb
from . import part_factory_alias as pfa
from . import part_factory_enrich as pfe
from . import assembly
from . import assembly_config
from . import assembly_factory_alias as afalias
from . import assembly_factory_assy as afa
from .render import render_cfg_merge


class Project(project_config.Configuration):

    class PartLock(object):
        def __init__(self, prj, part_name: str):
            if not part_name in prj.part_locks:
                prj.part_locks[part_name] = threading.Lock()
            self.lock = prj.part_locks[part_name]

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    class AssemblyLock(object):
        def __init__(self, prj, assembly_name: str):
            if not assembly_name in prj.assembly_locks:
                prj.assembly_locks[assembly_name] = threading.Lock()
            self.lock = prj.assembly_locks[assembly_name]

        def __enter__(self, *_args):
            self.lock.acquire()

        def __exit__(self, *_args):
            self.lock.release()

    def __init__(self, ctx, name, path):
        super().__init__(name, path)
        self.ctx = ctx

        # Protect the critical sections from access in different threads
        self.lock = threading.Lock()

        # The 'path' parameter is the config filename or the directory
        # where 'partcad.yaml' is present.
        # 'self.path' has to be set to the directory name.
        dir_name = path
        if not os.path.isdir(dir_name):
            dir_name = os.path.dirname(os.path.abspath(dir_name))
        self.path = dir_name

        # self.part_configs contains the configs of all the parts in this project
        if "parts" in self.config_obj and not self.config_obj["parts"] is None:
            self.part_configs = self.config_obj["parts"]
        else:
            self.part_configs = {}
        # self.parts contains all the initialized parts in this project
        self.parts = {}
        self.part_locks = {}

        # self.assembly_configs contains the configs of all the assemblies in this project
        if (
            "assemblies" in self.config_obj
            and not self.config_obj["assemblies"] is None
        ):
            self.assembly_configs = self.config_obj["assemblies"]
        else:
            self.assembly_configs = {}
        # self.assemblies contains all the initialized assemblies in this project
        self.assemblies = {}
        self.assembly_locks = {}

        if "desc" in self.config_obj and isinstance(
            self.config_obj["desc"], str
        ):
            self.desc = self.config_obj["desc"]
        else:
            self.desc = ""

        self.init_parts()
        self.init_assemblies()

    def get_part_config(self, part_name):
        if not part_name in self.part_configs:
            return None
        return self.part_configs[part_name]

    def init_parts(self):
        if self.part_configs is None:
            return

        for part_name in self.part_configs:
            config = self.get_part_config(part_name)
            config = part_config.PartConfiguration.normalize(part_name, config)
            self.init_part_by_config(config)

    def init_part_by_config(self, config, source_project=None):
        if source_project is None:
            source_project = self

        part_name: str = config["name"]

        if not "type" in config:
            raise Exception(
                "ERROR: Part type is not specified: %s: %s"
                % (part_name, config)
            )
        elif config["type"] == "cadquery":
            pfc.PartFactoryCadquery(self.ctx, source_project, self, config)
        elif config["type"] == "build123d":
            pfb.PartFactoryBuild123d(self.ctx, source_project, self, config)
        elif config["type"] == "step":
            pfs.PartFactoryStep(self.ctx, source_project, self, config)
        elif config["type"] == "stl":
            pfstl.PartFactoryStl(self.ctx, source_project, self, config)
        elif config["type"] == "3mf":
            pf3.PartFactory3mf(self.ctx, source_project, self, config)
        elif config["type"] == "scad":
            pfscad.PartFactoryScad(self.ctx, source_project, self, config)
        elif config["type"] == "alias":
            pfa.PartFactoryAlias(self.ctx, source_project, self, config)
        elif config["type"] == "enrich":
            pfe.PartFactoryEnrich(self.ctx, source_project, self, config)
        else:
            pc_logging.error(
                "Invalid part type encountered: %s: %s" % (part_name, config)
            )
            return None

        # Initialize aliases if they are declared implicitly
        if "aliases" in config and not config["aliases"] is None:
            for alias in config["aliases"]:
                if ";" in part_name:
                    # Copy parameters
                    alias += part_name[part_name.index(";") :]
                alias_part_config = {
                    "type": "alias",
                    "name": alias,
                    "source": ":" + part_name,
                }
                alias_part_config = part_config.PartConfiguration.normalize(
                    alias, alias_part_config
                )
                pfa.PartFactoryAlias(
                    self.ctx, source_project, self, alias_part_config
                )

    def get_part(self, part_name, func_params=None) -> part.Part:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: {str: typing.Any} = {}
        if ";" in part_name:
            has_name_params = True
            base_part_name = part_name.split(";")[0]
            part_name_params_string = part_name.split(";")[1]

            for kv in part_name_params_string.split[","]:
                k, v = kv.split("")
                params[k] = v
        else:
            has_name_params = False
            base_part_name = part_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = part_name
        else:
            # Determine the name we want this parameterized part to have
            result_name = base_part_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if result_name in self.parts and not self.parts[result_name] is None:
            p = self.parts[result_name]
            self.lock.release()
            return p

        with Project.PartLock(self, result_name):
            # Release the project lock, and continue with holding the part lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular part name, no params (part_name == result_name)
                if not part_name in self.part_configs:
                    # We don't know anything about such a part
                    pc_logging.error(
                        "Part '%s' not found in '%s'", part_name, self.name
                    )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_part_config(part_name)
                config = part_config.PartConfiguration.normalize(
                    part_name, config
                )
                self.init_part_by_config(config)

                if not part_name in self.parts or self.parts[part_name] is None:
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized part %s"
                        % part_name
                    )
                return self.parts[part_name]

            # This part has params (part_name != result_name)
            if not base_part_name in self.parts:
                pc_logging.error(
                    "Base part '%s' not found in '%s'",
                    base_part_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base part: %s" % base_part_name)

            # Now we have the original part name and the complete set of parameters
            config = self.part_configs[base_part_name]
            if config is None:
                pc_logging.error(
                    "The config for the base part '%s' is not found in '%s'",
                    base_part_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_part_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = part_config.PartConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_part_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the part
            pc_logging.debug(
                "Initializing a parametrized part: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized part using the following config: %s"
            #     % pformat(config)
            # )
            self.init_part_by_config(config)

            # See if it worked
            if not result_name in self.parts:
                pc_logging.error(
                    "Failed to instantiate parameterized part '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.parts[result_name]

    def get_assembly_config(self, assembly_name):
        if not assembly_name in self.assembly_configs:
            return None
        return self.assembly_configs[assembly_name]

    def init_assemblies(self):
        if self.assembly_configs is None:
            return

        for assembly_name in self.assembly_configs:
            config = self.get_assembly_config(assembly_name)
            config = assembly_config.AssemblyConfiguration.normalize(
                assembly_name, config
            )
            factory.instantiate("assembly", self.ctx, self, config)

    def get_assembly(
        self, assembly_name, func_params=None
    ) -> assembly.Assembly:
        if func_params is None or not func_params:
            has_func_params = False
        else:
            has_func_params = True

        params: {str: typing.Any} = {}
        if ";" in assembly_name:
            has_name_params = True
            base_assembly_name = assembly_name.split(";")[0]
            assembly_name_params_string = assembly_name.split(";")[1]

            for kv in assembly_name_params_string.split[","]:
                k, v = kv.split("")
                params[k] = v
        else:
            has_name_params = False
            base_assembly_name = assembly_name

        if has_func_params:
            params = {**params, **func_params}
            has_name_params = True

        if not has_name_params:
            result_name = assembly_name
        else:
            # Determine the name we want this parameterized assembly to have
            result_name = base_assembly_name + ";"
            result_name += ",".join(
                map(lambda n: n + "=" + str(params[n]), sorted(params))
            )

        self.lock.acquire()

        # See if it's already available
        if (
            result_name in self.assemblies
            and not self.assemblies[result_name] is None
        ):
            p = self.assemblies[result_name]
            self.lock.release()
            return p

        with Project.AssemblyLock(self, result_name):
            # Release the project lock, and continue with holding the part lock only
            self.lock.release()

            if not has_name_params:
                # This is just a regular assembly name, no params (assembly_name == result_name)
                if not assembly_name in self.assembly_configs:
                    # We don't know anything about such an assembly
                    pc_logging.error(
                        "Assembly '%s' not found in '%s'",
                        assembly_name,
                        self.name,
                    )
                    return None
                # This is not yet created (invalidated?)
                config = self.get_assembly_config(assembly_name)
                config = assembly_config.AssemblyConfiguration.normalize(
                    assembly_name, config
                )
                factory.instantiate("assembly", self.ctx, self, config)

                if (
                    not assembly_name in self.assemblies
                    or self.assemblies[assembly_name] is None
                ):
                    pc_logging.error(
                        "Failed to instantiate a non-parametrized assembly %s"
                        % assembly_name
                    )
                return self.assemblies[assembly_name]

            # This assembly has params (part_name != result_name)
            if not base_assembly_name in self.assemblies:
                pc_logging.error(
                    "Base assembly '%s' not found in '%s'",
                    base_assembly_name,
                    self.name,
                )
                return None
            pc_logging.debug("Found the base assembly: %s" % base_assembly_name)

            # Now we have the original assembly name and the complete set of parameters
            config = self.assembly_configs[base_assembly_name]
            if config is None:
                pc_logging.error(
                    "The config for the base assembly '%s' is not found in '%s'",
                    base_assembly_name,
                    self.name,
                )
                return None

            config = copy.deepcopy(config)
            if (
                not "parameters" in config or config["parameters"] is None
            ) and (config["type"] != "enrich"):
                pc_logging.error(
                    "Attempt to parametrize '%s' of '%s' which has no parameters: %s",
                    base_assembly_name,
                    self.name,
                    str(config),
                )
                return None

            # Expand the config object so that the parameter values can be set
            config = assembly_config.AssemblyConfiguration.normalize(
                result_name, config
            )
            config["orig_name"] = base_assembly_name

            # Fill in the parameter values
            param_name: str
            if "parameters" in config and not config["parameters"] is None:
                # Filling "parameters"
                for param_name, param_value in params.items():
                    if config["parameters"][param_name]["type"] == "string":
                        config["parameters"][param_name]["default"] = str(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "int":
                        config["parameters"][param_name]["default"] = int(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "float":
                        config["parameters"][param_name]["default"] = float(
                            param_value
                        )
                    elif config["parameters"][param_name]["type"] == "bool":
                        if isinstance(param_value, str):
                            if param_value.lower() == "true":
                                config["parameters"][param_name][
                                    "default"
                                ] = True
                            else:
                                config["parameters"][param_name][
                                    "default"
                                ] = False
                        else:
                            config["parameters"][param_name]["default"] = bool(
                                param_value
                            )
            else:
                # Filling "with"
                if not "with" in config:
                    config["with"] = {}
                for param_name, param_value in params.items():
                    config["with"][param_name] = param_value

            # Now initialize the assembly
            pc_logging.debug(
                "Initializing a parametrized assembly: %s" % result_name
            )
            # pc_logging.debug(
            #     "Initializing a parametrized assembly using the following config: %s"
            #     % pformat(config)
            # )
            factory.instantiate("assembly", self.ctx, self, config)

            # See if it worked
            if not result_name in self.assemblies:
                pc_logging.error(
                    "Failed to instantiate parameterized assembly '%s' in '%s'",
                    result_name,
                    self.name,
                )
                return None

            return self.assemblies[result_name]

    def add_import(self, alias, location):
        if ":" in location:
            location_param = "url"
            if location.endswith(".tar.gz"):
                location_type = "tar"
            else:
                location_type = "git"
        else:
            location_param = "path"
            location_type = "local"

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            config = yaml.load(fp)
            fp.close()

        for elem in config:
            if elem == "import":
                imports = config["import"]
                imports[alias] = {
                    location_param: location,
                    "type": location_type,
                }
                break  # no need to iterate further
        with open(self.config_path) as fp:
            yaml.dump(config, fp)
            fp.close()

    def _validate_path(self, path, extension) -> (bool, str, str):
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        root = self.config_dir
        if not os.path.isabs(root):
            root = os.path.abspath(root)

        if not path.startswith(root):
            pc_logging.error("Can't add files outside of the package")
            return False, None, None

        path = path[len(root) :]
        if path[0] == os.path.sep:
            path = path[1:]

        name = path
        if name.endswith(".%s" % extension):
            name = name[: -len(extension) - 1]

        return True, path, name

    def _add_component(
        self, kind: str, path: str, section: str, ext_by_kind
    ) -> bool:
        if kind in ext_by_kind:
            ext = ext_by_kind[kind]
        else:
            ext = kind

        valid, path, name = self._validate_path(path, ext)
        if not valid:
            return False

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        with open(self.config_path) as fp:
            config = yaml.load(fp)
            fp.close()

        obj = {"type": kind}
        if name == path:
            obj["path"] = path

        found = False
        for elem in config:
            if elem == section:
                config_section = config[section]
                if config_section is None:
                    config_section = {}
                config_section[name] = obj
                config[section] = config_section
                found = True
                break  # no need to iterate further
        if not found:
            config[section] = {name: obj}

        with open(self.config_path, "w") as fp:
            yaml.dump(config, fp)
            fp.close()

        return True

    def add_part(self, kind: str, path: str) -> bool:
        pc_logging.info("Adding the part %s of type %s" % (path, kind))
        ext_by_kind = {
            "cadquery": "py",
            "build123d": "py",
        }
        return self._add_component(kind, path, "parts", ext_by_kind)

    def add_assembly(self, kind: str, path: str) -> bool:
        pc_logging.info("Adding the assembly %s of type %s" % (path, kind))
        ext_by_kind = {}
        return self._add_component(kind, path, "assemblies", ext_by_kind)

    async def render_async(
        self, parts=None, assemblies=None, format=None, output_dir=None
    ):
        with pc_logging.Action("RenderPkg", self.name):
            # Override the default output_dir.
            # TODO(clairbee): pass the preference downstream without making a
            # persistent change.
            if not output_dir is None:
                self.config_obj["render"]["output_dir"] = output_dir

            if (
                "render" in self.config_obj
                and not self.config_obj["render"] is None
            ):
                render = self.config_obj["render"]
            else:
                render = {}

            # Enumerating all parts and assemblies
            if parts is None:
                parts = []
                if "parts" in self.config_obj:
                    parts = self.config_obj["parts"].keys()
            if assemblies is None:
                assemblies = []
                if "assemblies" in self.config_obj:
                    assemblies = self.config_obj["assemblies"].keys()

            # Render
            tasks = []
            for part_name in parts:
                part = self.get_part(part_name)
                if not part is None:
                    part_render = render
                    if (
                        "render" in part.config
                        and not part.config["render"] is None
                    ):
                        part_render = render_cfg_merge(
                            part_render, part.config["render"]
                        )

                    # Determine which formats need to be rendered.
                    # The format needs to be rendered either if it's mentioned in the config
                    # or if it's explicitly requested in the params (e.g. comes from the
                    # command line).
                    if format is None and "svg" in part_render:
                        render_svg = True
                    elif not format is None and format == "svg":
                        render_svg = True
                    else:
                        render_svg = False

                    if format is None and "png" in part_render:
                        render_png = True
                    elif not format is None and format == "png":
                        render_png = True
                    else:
                        render_png = False

                    if format is None and "step" in part_render:
                        render_step = True
                    elif not format is None and format == "step":
                        render_step = True
                    else:
                        render_step = False

                    if format is None and "stl" in part_render:
                        render_stl = True
                    elif not format is None and format == "stl":
                        render_stl = True
                    else:
                        render_stl = False

                    if format is None and "3mf" in part_render:
                        render_3mf = True
                    elif not format is None and format == "3mf":
                        render_3mf = True
                    else:
                        render_3mf = False

                    if format is None and "threejs" in part_render:
                        render_threejs = True
                    elif not format is None and format == "threejs":
                        render_threejs = True
                    else:
                        render_threejs = False

                    if format is None and "obj" in part_render:
                        render_obj = True
                    elif not format is None and format == "obj":
                        render_obj = True
                    else:
                        render_obj = False

                    if render_svg:
                        tasks.append(part.render_svg_async(self))
                    if render_png:
                        tasks.append(part.render_png_async(self))
                    if render_step:
                        tasks.append(part.render_step_async(self))
                    if render_stl:
                        tasks.append(part.render_stl_async(self))
                    if render_3mf:
                        tasks.append(part.render_3mf_async(self))
                    if render_threejs:
                        tasks.append(part.render_threejs_async(self))
                    if render_obj:
                        tasks.append(part.render_obj_async(self))

            for assembly_name in assemblies:
                assembly = self.get_assembly(assembly_name)
                if not assembly is None:
                    assy_render = render
                    if (
                        "render" in assembly.config
                        and not assembly.config["render"] is None
                    ):
                        assy_render = render_cfg_merge(
                            assy_render, assembly.config["render"]
                        )

                    # Determine which formats need to be rendered.
                    # The format needs to be rendered either if it's mentioned in the config
                    # or if it's explicitly requested in the params (e.g. comes from the
                    # command line).
                    if format is None and "svg" in assy_render:
                        render_svg = True
                    elif not format is None and format == "svg":
                        render_svg = True
                    else:
                        render_svg = False

                    if format is None and "png" in assy_render:
                        render_png = True
                    elif not format is None and format == "png":
                        render_png = True
                    else:
                        render_png = False

                    if format is None and "step" in assy_render:
                        render_step = True
                    elif not format is None and format == "step":
                        render_step = True
                    else:
                        render_step = False

                    if format is None and "stl" in assy_render:
                        render_stl = True
                    elif not format is None and format == "stl":
                        render_stl = True
                    else:
                        render_stl = False

                    if format is None and "3mf" in assy_render:
                        render_3mf = True
                    elif not format is None and format == "3mf":
                        render_3mf = True
                    else:
                        render_3mf = False

                    if format is None and "threejs" in assy_render:
                        render_threejs = True
                    elif not format is None and format == "threejs":
                        render_threejs = True
                    else:
                        render_threejs = False

                    if format is None and "obj" in assy_render:
                        render_obj = True
                    elif not format is None and format == "obj":
                        render_obj = True
                    else:
                        render_obj = False

                    if render_svg:
                        tasks.append(assembly.render_svg_async(self))
                    if render_png:
                        tasks.append(assembly.render_png_async(self))
                    if render_step:
                        tasks.append(assembly.render_step_async(self))
                    if render_stl:
                        tasks.append(assembly.render_stl_async(self))
                    if render_3mf:
                        tasks.append(assembly.render_3mf_async(self))
                    if render_threejs:
                        tasks.append(assembly.render_threejs_async(self))
                    if render_obj:
                        tasks.append(assembly.render_obj_async(self))

            await asyncio.gather(*tasks)

    def render(self, parts=None, assemblies=None, format=None, output_dir=None):
        asyncio.run(self.render_async(parts, assemblies, format, output_dir))



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.
#

from . import part_factory_file as pff


class PartFactoryPython(pff.PartFactoryFile):
    def __init__(self, ctx, source_project, target_project, part_config):
        super().__init__(
            ctx, source_project, target_project, part_config, extension=".py"
        )

        self.runtime = self.ctx.get_python_runtime(self.project.python_version)
        self.runtime.prepare_for_package(source_project)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-22
#
# Licensed under Apache License, Version 2.0.
#

# The alias of the current package: the one Context object was initialized with.
CURRENT = "."

# The top level package: the top-most found by Context while traversing '..'.
ROOT = "/"

# Which file to look for if the directory name is passed instead of the config
# filename.
DEFAULT_PACKAGE_CONFIG = "partcad.yaml"



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import cadquery as cq
import build123d as b3d

import asyncio
import os
import shutil
import tempfile

from .render import *
from .plugins import *
from .shape_config import ShapeConfiguration
from . import logging as pc_logging
from . import sync_threads as pc_thread


class Shape(ShapeConfiguration):
    name: str
    desc: str
    svg_path: str
    svg_url: str
    # shape: None | b3d.TopoDS_Shape | OCP.TopoDS.TopoDS_Solid

    def __init__(self, config):
        super().__init__(config)
        self.shape = None
        self.compound = None

        # Leave the svg path empty to get it created on demand
        self.svg_lock = asyncio.Lock()
        self.svg_path = None
        self.svg_url = None

    async def get_wrapped(self):
        shape = await self.get_shape()
        if "offset" in self.config:
            b3d_solid = b3d.Solid.make_box(1, 1, 1)
            b3d_solid.wrapped = shape
            b3d_solid.relocate(b3d.Location(*self.config["offset"]))
            shape = b3d_solid.wrapped
        if "scale" in self.config:
            b3d_solid = b3d.Solid.make_box(1, 1, 1)
            b3d_solid.wrapped = shape
            b3d_solid = b3d_solid.scale(self.config["scale"])
            shape = b3d_solid.wrapped
        return shape

    async def get_cadquery(self) -> cq.Shape:
        cq_solid = cq.Solid.makeBox(1, 1, 1)
        cq_solid.wrapped = await self.get_wrapped()
        return cq_solid

    async def get_build123d(self) -> b3d.Solid:
        b3d_solid = b3d.Solid.make_box(1, 1, 1)
        b3d_solid.wrapped = await self.get_wrapped()
        return b3d_solid

    async def show_async(self, show_object=None):
        with pc_logging.Action("Show", self.project_name, self.name):
            shape = await self.get_wrapped()
            if shape is not None:
                if show_object is None:
                    import importlib

                    ocp_vscode = importlib.import_module("ocp_vscode")
                    if ocp_vscode is None:
                        pc_logging.warning(
                            'Failed to load "ocp_vscode". Giving up on connection to VS Code.'
                        )
                    else:
                        try:
                            # ocp_vscode.config.status()
                            pc_logging.info(
                                'Visualizing in "OCP CAD Viewer"...'
                            )
                            # pc_logging.debug(self.shape)
                            ocp_vscode.show(
                                shape,
                                progress=None,
                                # TODO(clairbee): make showing (and the connection
                                # to ocp_vscode) a part of the context, and memorize
                                # which part was displayed last. Keep the camera
                                # if the part has not changed.
                                # reset_camera=ocp_vscode.Camera.KEEP,
                            )
                        except Exception as e:
                            pc_logging.warning(e)
                            pc_logging.warning(
                                'No VS Code or "OCP CAD Viewer" extension detected.'
                            )

                if show_object is not None:
                    show_object(
                        shape,
                        options={},
                    )

    def show(self, show_object=None):
        asyncio.run(self.show_async(show_object))

    async def render_svg_somewhere(self, project=None, filepath=None):
        """Renders an SVG file somewhere and ignore the project settings"""
        if filepath is None:
            filepath = tempfile.mktemp(".svg")

        cq_obj = await self.get_cadquery()

        opts = DEFAULT_RENDER_SVG_OPTS
        bb = cq_obj.BoundingBox()
        xlen = bb.xlen
        ylen = bb.ylen
        zlen = bb.zlen
        len = max(xlen, ylen, zlen)
        if len > 300.0:
            len = 300.0
        opts["strokeWidth"] = len / 150.0

        def do_render_svg():
            nonlocal cq_obj, filepath, opts
            cq_obj = cq_obj.rotate((0, 0, 0), (1, -1, 0.75), 180)
            cq.exporters.export(cq_obj, filepath, opt=opts)

        await pc_thread.run(do_render_svg)

        self.svg_path = filepath

    async def _get_svg_path(self, project):
        async with self.svg_lock:
            if self.svg_path is None:
                await self.render_svg_somewhere(project, None)
            return self.svg_path

    def render_getopts(
        self,
        kind,
        extension,
        project=None,
        filepath=None,
    ):
        if not project is None:
            render_opts = project.config_obj["render"]
        else:
            render_opts = {}

        if "render" in self.config and not self.config["render"] is None:
            render_opts = render_cfg_merge(render_opts, self.config["render"])

        if kind in render_opts and not render_opts[kind] is None:
            if isinstance(render_opts[kind], str):
                opts = {"prefix": render_opts[kind]}
            else:
                opts = render_opts[kind]
        else:
            opts = {}

        # Using the project's config defaults if any
        if filepath is None:
            if "path" in opts and not opts["path"] is None:
                filepath = opts["path"]
            elif "prefix" in opts and not opts["prefix"] is None:
                filepath = opts["prefix"]
            else:
                filepath = "."

            # Check if the format specific section of the config (e.g. "png")
            # provides a relative path and there is output dir in cmd line or
            # the generic section of rendering options in the config.
            if not os.path.isabs(filepath):
                if "output_dir" in render_opts:
                    filepath = os.path.join(render_opts["output_dir"], filepath)
                elif not project is None:
                    filepath = os.path.join(project.config_dir, filepath)

            if os.path.isdir(filepath):
                filepath = os.path.join(filepath, self.name + extension)

        pc_logging.debug("Rendering: %s" % filepath)

        return opts, filepath

    async def render_svg_async(
        self,
        project=None,
        filepath=None,
    ):
        with pc_logging.Action("RenderSVG", self.project_name, self.name):
            _, filepath = self.render_getopts("svg", ".svg", project, filepath)

            # This creates a temporary file, but it allows to reuse the file
            # with other consumers of self._get_svg_path()
            svg_path = await self._get_svg_path(project)
            shutil.copyfile(svg_path, filepath)

    def render_svg(
        self,
        project=None,
        filepath=None,
    ):
        asyncio.run(self.render_svg_async(project, filepath))

    async def render_png_async(
        self,
        project=None,
        filepath=None,
        width=None,
        height=None,
    ):
        with pc_logging.Action("RenderPNG", self.project_name, self.name):
            if not plugins.export_png.is_supported():
                pc_logging.error("Export to PNG is not supported")
                return

            png_opts, filepath = self.render_getopts(
                "png", ".png", project, filepath
            )

            if width is None:
                if "width" in png_opts and not png_opts["width"] is None:
                    width = png_opts["width"]
                else:
                    width = DEFAULT_RENDER_WIDTH
            if height is None:
                if "height" in png_opts and not png_opts["height"] is None:
                    height = png_opts["height"]
                else:
                    height = DEFAULT_RENDER_HEIGHT

            # Render the vector image
            svg_path = await self._get_svg_path(project)

            def do_render_png():
                nonlocal project, svg_path, width, height, filepath
                plugins.export_png.export(
                    project, svg_path, width, height, filepath
                )

            await pc_thread.run(do_render_png)

    def render_png(
        self,
        project=None,
        filepath=None,
        width=None,
        height=None,
    ):
        asyncio.run(self.render_png_async(project, filepath, width, height))

    async def render_step_async(
        self,
        project=None,
        filepath=None,
    ):
        with pc_logging.Action("RenderSTEP", self.project_name, self.name):
            step_opts, filepath = self.render_getopts(
                "step", ".step", project, filepath
            )

            cq_obj = await self.get_cadquery()

            def do_render_step():
                nonlocal project, filepath, cq_obj
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(cq_obj, filepath)

            await pc_thread.run(do_render_step)

    def render_step(
        self,
        project=None,
        filepath=None,
    ):
        asyncio.run(self.render_step_async(project, filepath))

    async def render_stl_async(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderSTL", self.project_name, self.name):
            stl_opts, filepath = self.render_getopts(
                "stl", ".stl", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in stl_opts
                    and not stl_opts["tolerance"] is None
                ):
                    tolerance = stl_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in stl_opts
                    and not stl_opts["angularTolerance"] is None
                ):
                    angularTolerance = stl_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_stl():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                )

            await pc_thread.run(do_render_stl)

    def render_stl(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_stl_async(
                project, filepath, tolerance, angularTolerance
            )
        )

    async def render_3mf_async(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("Render3MF", self.project_name, self.name):
            threemf_opts, filepath = self.render_getopts(
                "3mf", ".3mf", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in threemf_opts
                    and not threemf_opts["tolerance"] is None
                ):
                    tolerance = threemf_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in threemf_opts
                    and not threemf_opts["angularTolerance"] is None
                ):
                    angularTolerance = threemf_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_3mf():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                )

            await pc_thread.run(do_render_3mf)

    def render_3mf(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_3mf_async(
                project, filepath, tolerance, angularTolerance
            )
        )

    async def render_threejs_async(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderThreeJS", self.project_name, self.name):
            threejs_opts, filepath = self.render_getopts(
                "threejs", ".json", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in threejs_opts
                    and not threejs_opts["tolerance"] is None
                ):
                    tolerance = threejs_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in threejs_opts
                    and not threejs_opts["angularTolerance"] is None
                ):
                    angularTolerance = threejs_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_threejs():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                    exportType=cq.exporters.ExportTypes.TJS,
                )

            await pc_thread.run(do_render_threejs)

    def render_threejs(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_threejs_async(
                project, filepath, tolerance, angularTolerance
            )
        )

    async def render_obj_async(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderOBJ", self.project_name, self.name):
            obj_opts, filepath = self.render_getopts(
                "obj", ".obj", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in obj_opts
                    and not obj_opts["tolerance"] is None
                ):
                    tolerance = obj_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in obj_opts
                    and not obj_opts["angularTolerance"] is None
                ):
                    angularTolerance = obj_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_obj():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                try:
                    vertices, triangles = cq_obj.tessellate(
                        tolerance, angularTolerance
                    )

                    with open(filepath, "w") as f:
                        f.write("# OBJ file\n")
                        for v in vertices:
                            f.write("v %.4f %.4f %.4f\n" % (v.x, v.y, v.z))
                        for p in triangles:
                            f.write("f")
                            for i in p:
                                f.write(" %d" % (i + 1))
                            f.write("\n")
                except:
                    pc_logging.error("Exception while exporting to " + filepath)

            await pc_thread.run(do_render_obj)

    def render_obj(
        self,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_obj_async(
                project, filepath, tolerance, angularTolerance
            )
        )

    async def render_txt_async(self, project=None, filepath=None):
        with pc_logging.Action("RenderTXT", self.project_name, self.name):
            if filepath is None:
                filepath = self.path + "/bom.txt"

            if not project is None:
                project.ctx.ensure_dirs_for_file(filepath)
            file = open(filepath, "w+")
            file.write("BoM:\n")
            await self._render_txt_real(file)
            file.close()

    def render_txt(self, project=None, filepath=None):
        asyncio.run(self.render_txt_async(project, filepath))

    async def render_markdown_async(self, project=None, filepath=None):
        with pc_logging.Action("RenderMD", self.project_name, self.name):
            if filepath is None:
                filepath = self.path + "/README.md"

            bom_file = open(filepath, "w+")
            bom_file.write(
                "# "
                + self.name
                + "\n"
                + "## Bill of Materials\n"
                + "| Part | Count* | Vendor | SKU | Preview |\n"
                + "| -- | -- | -- | -- | -- |\n"
            )
            self._render_markdown_real(bom_file)
            bom_file.write(
                """
(\\*) The `Count` field is the number of SKUs to be ordered.
It already takes into account the number of items per SKU.
            """
            )
            bom_file.close()

    def render_markdown(self, project=None, filepath=None):
        asyncio.run(self.render_markdown_async(project, filepath))



from build123d import Location

from .globals import (
    init,
    fini,
    create_package,
    get_part,
    get_part_cadquery,
    get_part_build123d,
    get_assembly,
    get_assembly_cadquery,
    get_assembly_build123d,
    _partcad_context,
    render,
)
from .consts import *
from .context import Context
from .assembly import Assembly
from .part import Part
from .project_factory_local import ProjectFactoryLocal
from .project_factory_git import ProjectFactoryGit
from .project_factory_tar import ProjectFactoryTar
from .user_config import user_config
from .plugins import plugins
from .plugin_export_png_reportlab import PluginExportPngReportlab
from .logging_ansi_terminal import init as logging_ansi_terminal_init
from .logging_ansi_terminal import fini as logging_ansi_terminal_fini
from . import logging
from . import utils

__all__ = [
    "config",
    "context",
    "shape",
    "part",
    "part_factory",
    "part_factory_step",
    "part_factory_cadquery",
    "project",
    "project_factory",
    "project_factory_local",
    "project_factory_git",
    "project_factory_tar",
    "assembly",
    "assembly_factory",
    "assembly_factory_python",
    "scene",
    "main_cli",
    "plugins",
]

__version__: str = "0.5.10"



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os
import shutil
import subprocess

from . import runtime_python


class PyPyPythonRuntime(runtime_python.PythonRuntime):
    def __init__(self, ctx, version=None):
        super().__init__(ctx, "pypy", version)

        if not self.initialized:
            if shutil.which("pypy") is None:
                raise Exception(
                    "ERROR: PartCAD is configured to use missing pypy to execute Python scripts (CadQuery, build123d etc)"
                )

            os.makedirs(self.path)
            try:
                subprocess.run(
                    ["conda", "create", "-p", self.path, "pypy", "python=%s" % version]
                )
                subprocess.run(["conda", "install", "-p", self.path, "scipy"])
            except Exception as e:
                shutil.rmtree(self.path)
                raise e

    def run(self, cmd, stdin=""):
        return super().run(
            ["conda", "run", "--no-capture-output", "-p", self.path, "pypy"] + cmd,
            stdin,
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import build123d as b3d
from jinja2 import Environment, FileSystemLoader
import os
import yaml

from .assembly import Assembly, AssemblyChild
from . import assembly_factory_file as aff
from . import logging as pc_logging
from .utils import resolve_resource_path


class AssemblyFactoryAssy(aff.AssemblyFactoryFile):
    def __init__(self, ctx, project, assembly_config):
        with pc_logging.Action(
            "InitASSY", project.name, assembly_config["name"]
        ):
            super().__init__(ctx, project, assembly_config, extension=".assy")
            # Complement the config object here if necessary
            self._create(assembly_config)

    def instantiate(self, assembly):
        # # This method is best executed on a thread but the current Python version
        # # might not be good enough to do that.
        # try:
        #     # Try running on the current event loop if any.
        #     loop = asyncio.get_running_loop()
        #     # task = loop.create_task(self.instantiate_async(assembly))
        #     # task.
        #     f = asyncio.run_coroutine_threadsafe(self.instantiate_async(assembly), loop)
        #     raise Exception("IT WORKS")
        # except RuntimeError as e:
        #     print(e)
        #     # Running on a dedicated thread and there is no event loop here yet.
        asyncio.run(self.instantiate_async(assembly))

    async def instantiate_async(self, assembly):
        with pc_logging.Action("ASSY", assembly.project_name, assembly.name):
            assy = {}
            if os.path.exists(self.path):
                params = {}
                if "parameters" in self.assembly_config:
                    for param_name, param in self.assembly_config[
                        "parameters"
                    ].items():
                        params["param_" + param_name] = param["default"]
                params["name"] = self.assembly_config["name"]

                # Read the body of the configuration file
                fp = open(self.path, "r")
                config = fp.read()
                fp.close()

                # Resolve Jinja templates
                template = Environment(
                    loader=FileSystemLoader(
                        os.path.dirname(self.path) + os.path.sep
                    )
                ).from_string(config)
                config = template.render(params)

                # Parse the resulting config
                try:
                    assy = yaml.safe_load(config)
                except Exception as e:
                    pc_logging.error(
                        "ERROR: Failed to parse the assembly file %s"
                        % self.path
                    )
                if assy is None:
                    assy = {}
            else:
                pc_logging.error(
                    "ERROR: Assembly file not found: %s" % self.path
                )

            result = await self.handle_node(assembly, assy)
            if not result is None:
                assembly.children.append(
                    AssemblyChild(result[0], result[1], result[2])
                )
                # Keep part reference counter for bill-of-materials purposes
                result[0].ref_inc()
            else:
                pc_logging.warning("Assembly is empty")

            self.ctx.stats_assemblies_instantiated += 1

    async def handle_node_list(self, assembly, node_list):
        tasks = []
        for link in node_list:
            tasks.append(self.handle_node(assembly, link))

        # TODO(clairbee): revisit whether non-determinism is acceptable here
        for f in asyncio.as_completed(tasks):
            result = await f
            if not result is None:
                assembly.children.append(
                    AssemblyChild(result[0], result[1], result[2])
                )
                # Keep part reference counter for bill-of-materials purposes
                result[0].ref_inc()

    async def handle_node(self, assembly, node):
        # "name" is an optional parameter for both parts and assemblies
        if "name" in node:
            name = node["name"]
        else:
            name = None

        # "location" is an optional parameter for both parts and assemblies
        if "location" in node:
            l = node["location"]
            location = b3d.Location(
                (l[0][0], l[0][1], l[0][2]), (l[1][0], l[1][1], l[1][2]), l[2]
            )
        else:
            location = b3d.Location((0, 0, 0), (0, 0, 1), 0)

        # Check if this node is for an assembly
        if "links" in node and not node["links"] is None:
            item = Assembly(
                {"name": name}
            )  # TODO(clairbee): revisit why node["links"]) was used there
            item.instantiate = lambda x: True
            await self.handle_node_list(item, node["links"])
        else:
            # This is a node for a part or an assembly
            params = {}
            if "params" in node:
                for paramName in node["params"]:
                    params[paramName] = node["params"][paramName]

            if "assembly" in node:
                assy_name = node["assembly"]
                if "package" in node:
                    assy_name = node["package"] + ":" + assy_name
                elif not ":" in assy_name:
                    assy_name = ":" + assy_name
                item = self.ctx._get_assembly(assy_name, params)
                if item is None:
                    pc_logging.error("Assembly not found: %s" % name)
                    raise Exception("Assembly not found")
            elif "part" in node:
                part_name = node["part"]
                if "package" in node:
                    part_name = node["package"] + ":" + part_name
                elif not ":" in part_name:
                    part_name = ":" + part_name
                item = self.ctx._get_part(part_name, params)
                if item is None:
                    pc_logging.error("Part not found: %s" % name)
                    raise Exception("Part not found")
            else:
                item = None

        if not item is None:
            return [item, name, location]
        else:
            return None



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import os
from . import project_factory as pf


class LocalImportConfiguration:
    def __init__(self):
        self.import_config_path = self.config_obj.get("path")
        self.maybe_empty = False
        if "maybeEmpty" in self.config_obj:
            self.maybe_empty = self.config_obj.get("maybeEmpty")


class ProjectFactoryLocal(pf.ProjectFactory, LocalImportConfiguration):
    def __init__(self, ctx, parent, config):
        pf.ProjectFactory.__init__(self, ctx, parent, config)
        LocalImportConfiguration.__init__(self)

        if not os.path.isabs(self.import_config_path) and self.config_dir != "":
            self.import_config_path = os.path.join(
                self.config_dir, self.import_config_path
            )

        self.path = self.import_config_path

        if not self.maybe_empty:
            if not os.path.exists(self.import_config_path):
                raise Exception(
                    "PartCAD config not found: %s" % self.import_config_path
                )

        # Complement the config object here if necessary
        self._create(config)

        # TODO(clairbee): consider installing a symlink in the parent's project

        self._save()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-02-04
#
# Licensed under Apache License, Version 2.0.

import asyncio
import atexit
import logging
from logging.handlers import QueueHandler, QueueListener
import queue
import sys
import time
from typing import Any
import threading

from .logging import ops


def ansi_process_start(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "process_start: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "process_start",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_process_end(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "process_end: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "process_end",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_action_start(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "action_start: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "action_start",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_action_end(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "action_end: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "action_end",
            "op": op,
            "package": package,
            "item": item,
        },
    )


class AnsiTerminalProgressHandler(logging.Handler):
    MAX_LINES = 8

    def __init__(self, stream=sys.stdout) -> None:
        super().__init__()
        self.thread_lock = threading.Lock()
        if not stream is None:
            self.stream = stream
        else:
            self.stream = sys.stdout

        self.thread = None

        self.process = None
        self.process_target = None
        self.process_start = None
        self.footer_size = 0

        self.actions = {}
        self.actions_running = 0
        self.actions_total = 0

        self.last_output = 0.0

    def clear_footer(self):
        return "\u001b[1A\u001b[2K" * self.footer_size

    def run_thread(self):
        while not self.process is None:
            time.sleep(0.25)
            if time.time() - self.last_output > 0.5:
                self.emit(None)

    def emit(self, record):
        now = time.time()

        # output accumulates the string that will be emitted before 'return'
        output = ""

        # protect the status data store in 'self' from other threads
        self.thread_lock.acquire()

        if not record is None:
            ignore_message = False
            if hasattr(record, "pc_event"):
                if record.pc_event == "process_start":
                    self.process = record.op
                    self.process_target = record.package
                    if record.item is not None:
                        self.process_target += ":" + record.item
                    self.process_start = time.time()
                    self.actions_total = 0

                    if not self.thread is None:
                        self.thread_lock.release()
                        raise Exception("nested processes")
                    self.thread = threading.Thread(
                        target=self.run_thread,
                        name="ansi-terminal-update-" + str(now),
                    )
                    self.thread_lock.release()
                    self.thread.start()
                    return
                elif record.pc_event == "process_end":
                    self.process = None  # This also signals the thread to stop
                    output += self.clear_footer()
                    self.footer_size = 0

                    self.last_output = 0.0
                    th = self.thread
                    self.thread = None
                    self.thread_lock.release()

                    th.join()

                    if len(output) > 0:
                        print(output, end="", file=self.stream, flush=True)
                        self.stream.flush()
                    return

                elif record.pc_event == "action_start":
                    action = {"op": record.op}
                    target = record.package
                    if record.item is not None:
                        target += ":" + record.item
                    action["target"] = target
                    action["start"] = time.time()
                    self.actions[record.op + "-" + target] = action

                    self.actions_running += 1
                    self.actions_total += 1
                elif record.pc_event == "action_end":
                    target = record.package
                    if record.item is not None:
                        target += ":" + record.item
                    del self.actions[record.op + "-" + target]

                    self.actions_running -= 1
                ignore_message = True

        output += self.clear_footer()

        if not record is None:
            if not ignore_message:
                if record.levelno == logging.DEBUG:
                    output += "\033[94mDEBUG:\033[0m "
                elif record.levelno == logging.INFO:
                    output += "\033[92m\033[1mINFO:\033[0m "
                elif record.levelno == logging.WARN:
                    output += "\033[93mWARN:\033[0m "
                elif record.levelno == logging.ERROR:
                    output += "\033[91mERROR:\033[0m "
                elif record.levelno == logging.CRITICAL:
                    output += "\033[91m\033[1mCRIT:\033[0m "

                msg = self.format(record)
                output += "%s\n" % msg

        if not self.process is None:
            seconds = int(now - self.process_start)

            output += (
                "\033[92m\033[1m[ %d / %d ]\033[0m \033[1m%s\033[0m %s; %ds\n"
                % (
                    self.actions_running,
                    self.actions_total,
                    self.process,
                    self.process_target,
                    seconds,
                )
            )
            self.footer_size = 1

            sorted_actions = sorted(
                self.actions.values(), key=lambda a: a["start"]
            )
            sorted_actions = sorted_actions[: self.MAX_LINES]
            for action in sorted_actions:
                output += "\t\033[1m[%s]\033[0m %s [%ds]\n" % (
                    action["op"],
                    action["target"],
                    now - action["start"],
                )
                self.footer_size += 1
        else:
            self.footer_size = 0

        self.last_output = now
        self.thread_lock.release()
        if len(output) > 0:
            print(output, end="", file=self.stream, flush=True)

        return


listener = None
queue_handler: AnsiTerminalProgressHandler = None


def fini():
    global listener
    if not listener is None:
        listener.stop()
        listener = None


def init(stream=None):
    global listener
    global ops

    log_queue = queue.Queue(-1)
    queue_handler = QueueHandler(log_queue)
    output_handler = AnsiTerminalProgressHandler(stream=stream)

    logger = logging.getLogger("partcad")
    handlers = logger.handlers
    for handler in handlers:
        logger.removeHandler(handler)
    logger.addHandler(queue_handler)

    listener = QueueListener(log_queue, output_handler)
    listener.start()
    atexit.register(fini)

    ops.process_start = ansi_process_start
    ops.process_end = ansi_process_end
    ops.action_start = ansi_action_start
    ops.action_end = ansi_action_end



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import build123d as b3d

from . import part_factory_file as pff
from . import logging as pc_logging


class PartFactoryStl(pff.PartFactoryFile):
    def __init__(self, ctx, source_project, target_project, part_config):
        with pc_logging.Action(
            "InitSTL", target_project.name, part_config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                part_config,
                extension=".stl",
            )
            # Complement the config object here if necessary
            self._create(part_config)

    def instantiate(self, part):
        with pc_logging.Action("STL", part.project_name, part.name):
            shape = b3d.Mesher().read(self.path)[0].wrapped

            self.ctx.stats_parts_instantiated += 1

            return shape



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-09-30
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import assembly


# TODO(clairbee): introduce ShapeFactory
class AssemblyFactory:
    # TODO(clairbee): Make the next line work for assembly_factory_file only
    path: typing.Optional[str] = None

    def __init__(self, ctx, project, assembly_config, extension=""):
        self.ctx = ctx
        self.project = project
        self.assembly_config = assembly_config
        self.name = assembly_config["name"]
        self.orig_name = assembly_config["orig_name"]

    def _create(self, assembly_config):
        self.assembly = assembly.Assembly(assembly_config)
        self.assembly.project_name = (
            self.project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for assembly_factory_file only
        if self.path:
            self.assembly.path = self.path
        self.project.assemblies[self.name] = self.assembly

        self.assembly.instantiate = lambda assembly_self: self.instantiate(
            assembly_self
        )

        self.ctx.stats_assemblies += 1



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from git import Repo
import hashlib
import os
import pathlib
import time

from . import project_factory as pf
from . import logging as pc_logging
from .user_config import user_config


class GitImportConfiguration:
    def __init__(self):
        self.import_config_url = self.config_obj.get("url")
        self.import_revision = self.config_obj.get("revision")
        self.import_rel_path = self.config_obj.get("relPath")


class ProjectFactoryGit(pf.ProjectFactory, GitImportConfiguration):
    def __init__(self, ctx, parent, config):
        pf.ProjectFactory.__init__(self, ctx, parent, config)
        GitImportConfiguration.__init__(self)

        self.path = self._clone_or_update_repo(self.import_config_url)

        # Complement the config object here if necessary
        self._create(config)

        # TODO(clairbee): actually fill in the self.project object here

        self._save()

    def _clone_or_update_repo(self, repo_url, cache_dir=None):
        """
        Clones a Git repository to a local directory and keeps it up-to-date.

        Args:
          repo_url: URL of the Git repository to clone.
          cache_dir: Directory to store the cached copies of repositories (defaults to ".cache").

        Returns:
          Local path to the cloned repository.
        """

        if cache_dir is None:
            cache_dir = os.path.join(user_config.internal_state_dir, "git")

        # Generate a unique identifier for the repository based on its URL.
        repo_hash = hashlib.sha256(repo_url.encode()).hexdigest()
        if self.import_revision is not None:
            repo_hash += "-" + self.import_revision
        cache_path = os.path.join(cache_dir, repo_hash)
        guard_path = os.path.join(cache_path, ".partcad.git.cloned")

        # Check if the repository is already cached.
        if os.path.exists(cache_path):
            try:
                before = None

                # Try to open the existing repository and update it.
                if self.import_revision is None:
                    if user_config.force_update or (
                        time.time() - os.path.getmtime(guard_path) > 24 * 3600
                    ):
                        repo = Repo(cache_path)
                        origin = repo.remote("origin")
                        before = repo.active_branch.commit
                        origin.pull()
                        pathlib.Path(guard_path).touch()
                else:
                    repo = Repo(cache_path)
                    origin = repo.remote("origin")
                    before = repo.active_branch.commit
                    if user_config.force_update or (
                        before != self.import_revision
                        or (
                            time.time() - os.path.getmtime(guard_path)
                            > 24 * 3600
                        )
                    ):
                        # Need to check for updates
                        origin.fetch()
                        repo.git.checkout(self.import_revision, force=True)
                        pathlib.Path(guard_path).touch()

                if not before is None:
                    # Update was performed
                    after = repo.active_branch.commit
                    if before != after:
                        pc_logging.info(
                            "Updated the GIT repo: %s" % self.import_config_url
                        )
            except Exception as e:
                pc_logging.error("Exception: %s" % e)
                # Fall back to using the previous copy
        else:
            # Clone the repository if not cached.
            try:
                pc_logging.info(
                    "Cloning the GIT repo: %s" % self.import_config_url
                )
                repo = Repo.clone_from(repo_url, cache_path)
                if not self.import_revision is None:
                    repo.git.checkout(self.import_revision, force=True)

                if not os.path.exists(guard_path):
                    pathlib.Path(guard_path).touch()
            except Exception as e:
                raise RuntimeError(f"Failed to clone repository: {e}")

        if not self.import_rel_path is None:
            cache_path = os.path.join(cache_path, self.import_rel_path)

        return cache_path



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.

import os
import shutil
import threading

from .context import Context
from .assembly import Assembly
from .assembly_factory_assy import AssemblyFactoryAssy
from .assembly_factory_alias import AssemblyFactoryAlias
from .part import Part
from . import consts
from . import factory
from . import logging as pc_logging

global _partcad_context
_partcad_context = None
_partcad_context_lock = threading.Lock()

factory.register("assembly", "assy", AssemblyFactoryAssy)
factory.register("assembly", "alias", AssemblyFactoryAlias)


def init(config_path=None, search_root=True) -> Context:
    """Initialize the default context explicitly using the desired path."""
    global _partcad_context
    global _partcad_context_path
    global _partcad_context_lock

    with _partcad_context_lock:
        if _partcad_context is None:
            _partcad_context_path = config_path
            _partcad_context = Context(config_path, search_root=search_root)
            return _partcad_context

    if _partcad_context_path == config_path:
        return _partcad_context

    return Context(config_path, search_root=search_root)


def fini():
    global _partcad_context
    global _partcad_context_path
    global _partcad_context_lock

    with _partcad_context_lock:
        _partcad_context = None
        _partcad_context_path = None


def get_assembly(assembly_name, params=None) -> Assembly:
    """Get the assembly from the given project"""
    return init().get_assembly(assembly_name, params=params)


def get_assembly_cadquery(assembly_name, params=None) -> Assembly:
    """Get the assembly from the given project"""
    return init().get_assembly_cadquery(assembly_name, params=params)


def get_assembly_build123d(assembly_name, params=None) -> Assembly:
    """Get the assembly from the given project"""
    return init().get_assembly_build123d(assembly_name, params=params)


def get_part(part_name, params=None) -> Part:
    """Get the part from the given project"""
    return init().get_part(part_name, params=params)


def get_part_cadquery(part_name, params=None) -> Part:
    """Get the part from the given project"""
    return init().get_part_cadquery(part_name, params=params)


def get_part_build123d(part_name, params=None) -> Part:
    """Get the part from the given project"""
    return init().get_part_build123d(part_name, params=params)


def render(format=None, output_dir=None):
    return init().render(format, output_dir)


def create_package(dst_path=consts.DEFAULT_PACKAGE_CONFIG, private=False):
    if private:
        template_name = "init-private.yaml"
    else:
        template_name = "init-public.yaml"
    src_path = os.path.join(
        os.path.dirname(__file__), "template", template_name
    )

    if os.path.exists(dst_path):
        pc_logging.error("File already exists: %s" % dst_path)
        return False
    shutil.copyfile(src_path, dst_path)

    return True



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#

from .shape_config import ShapeConfiguration


class AssemblyConfiguration(ShapeConfiguration):
    def __init__(self, name, config=None):
        super().__init__(name, config)

    @staticmethod
    def normalize(name, config):
        if isinstance(config, str):
            # This is a short form alias
            config = {"type": "alias", "source": config}

        return ShapeConfiguration.normalize(name, config)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.

from .plugin_export_png import plugin_export_png


class Plugins:
    export_png = plugin_export_png


plugins = Plugins()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import os
import shutil
import subprocess
import tempfile

import build123d as b3d

from . import part_factory_file as pff
from . import logging as pc_logging


class PartFactoryScad(pff.PartFactoryFile):
    def __init__(self, ctx, source_project, target_project, part_config):
        with pc_logging.Action(
            "InitOpenSCAD", target_project.name, part_config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                part_config,
                extension=".scad",
            )
            # Complement the config object here if necessary
            self._create(part_config)

            self.project_dir = source_project.config_dir

    def instantiate(self, part):
        with pc_logging.Action("OpenSCAD", part.project_name, part.name):
            scad_path = shutil.which("openscad")
            if scad_path is None:
                raise Exception(
                    "OpenSCAD executable is not found. Please, install OpenSCAD first."
                )

            stl_path = tempfile.mktemp(".stl")
            p = subprocess.run(
                [
                    scad_path,
                    "--export-format",
                    "binstl",
                    "-o",
                    stl_path,
                    self.path,
                ],
                capture_output=True,
                # TODO(clairbee): cwd=self.project_dir,
            )

            shape = b3d.Mesher().read(stl_path)[0].wrapped
            os.unlink(stl_path)

            self.ctx.stats_parts_instantiated += 1

            return shape



# Original code taken from https://gist.github.com/SDI8/3137ee70649e4901913c7c8e6b534ec8

"""
MIT License
Copyright (c) 2022 Simon Dibbern
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import copyreg
from io import BytesIO

import cadquery as cq
import OCP


def _inflate_shape(data: bytes):
    with BytesIO(data) as bio:
        return cq.Shape.importBrep(bio)


def _reduce_shape(shape: cq.Shape):
    with BytesIO() as stream:
        shape.exportBrep(stream)
        return _inflate_shape, (stream.getvalue(),)


def _inflate_transform(*values: float):
    trsf = OCP.gp.gp_Trsf()
    trsf.SetValues(*values)
    return trsf


def _reduce_transform(transform: OCP.gp.gp_Trsf):
    return _inflate_transform, tuple(
        transform.Value(i, j) for i in range(1, 4) for j in range(1, 5)
    )


def _inflate_Gtransform(*values):
    gtrsf = OCP.gp.gp_GTrsf(values[0], values[1])
    return gtrsf


def _reduce_Gtransform(transform: OCP.gp.gp_GTrsf):
    return _inflate_Gtransform, tuple(
        [
            transform.VectorialPart(),
            transform.TranslationPart(),
        ]
    )


def _inflate_mat(*values):
    trsf = OCP.gp.gp_Mat(values[0], values[1], values[2])
    return trsf


def _reduce_mat(mat: OCP.gp.gp_Mat):
    return _inflate_mat, (
        mat.Column(1),
        mat.Column(2),
        mat.Column(3),
    )


def _inflate_compound(data: bytes):
    with BytesIO(data) as bio:
        shape = OCP.TopoDS.TopoDS_Compound()
        builder = OCP.BRep.BRep_Builder()
        OCP.BRepTools.BRepTools.Read_s(shape, bio, builder)
        return shape


def _reduce_compound(compound: OCP.TopoDS.TopoDS_Compound):
    with BytesIO() as stream:
        OCP.BRepTools.BRepTools.Write_s(compound, stream)
        return _inflate_compound, (stream.getvalue(),)


def _inflate_solid(data: bytes):
    with BytesIO(data) as bio:
        shape = OCP.TopoDS.TopoDS_Solid()
        builder = OCP.BRep.BRep_Builder()
        OCP.BRepTools.BRepTools.Read_s(shape, bio, builder)
        return shape


def _reduce_solid(compound: OCP.TopoDS.TopoDS_Solid):
    with BytesIO() as stream:
        OCP.BRepTools.BRepTools.Write_s(compound, stream)
        return _inflate_solid, (stream.getvalue(),)


def _inflate_shell(data: bytes):
    with BytesIO(data) as bio:
        shape = OCP.TopoDS.TopoDS_Shell()
        builder = OCP.BRep.BRep_Builder()
        OCP.BRepTools.BRepTools.Read_s(shape, bio, builder)
        return shape


def _reduce_shell(compound: OCP.TopoDS.TopoDS_Shell):
    with BytesIO() as stream:
        OCP.BRepTools.BRepTools.Write_s(compound, stream)
        return _inflate_shell, (stream.getvalue(),)


def _inflate_ax3(*values: float):
    ax3 = OCP.gp.gp_Ax3()
    ax3.SetLocation(values[0])
    ax3.SetDirection(values[1])
    return ax3


def _reduce_ax3(ax3: OCP.gp.gp_Ax3):
    return _inflate_ax3, (ax3.Location(), ax3.Direction())


def _inflate_pnt(*values: float):
    pnt = OCP.gp.gp_Pnt(values[0], values[1], values[2])
    return pnt


def _reduce_pnt(pnt: OCP.gp.gp_Pnt):
    return _inflate_pnt, (pnt.X(), pnt.Y(), pnt.Z())


def _inflate_dir(*values: float):
    dir = OCP.gp.gp_Dir(values[0], values[1], values[2])
    return dir


def _reduce_dir(dir: OCP.gp.gp_Dir):
    return _inflate_dir, (dir.X(), dir.Y(), dir.Z())


def _inflate_xyz(*values: float):
    dir = OCP.gp.gp_XYZ(values[0], values[1], values[2])
    return dir


def _reduce_xyz(dir: OCP.gp.gp_XYZ):
    return _inflate_xyz, (dir.X(), dir.Y(), dir.Z())


def register():
    """
    Registers pickle support functions for common CadQuery and OCCT objects.
    """

    for cls in (
        cq.Edge,
        cq.Compound,
        cq.Shell,
        cq.Face,
        cq.Solid,
        cq.Vertex,
        cq.Wire,
    ):
        copyreg.pickle(cls, _reduce_shape)

    copyreg.pickle(OCP.gp.gp_Pnt, _reduce_pnt)
    copyreg.pickle(OCP.gp.gp_Dir, _reduce_dir)
    copyreg.pickle(OCP.gp.gp_Mat, _reduce_mat)
    copyreg.pickle(OCP.gp.gp_XYZ, _reduce_xyz)

    copyreg.pickle(OCP.gp.gp_Ax3, _reduce_ax3)

    copyreg.pickle(cq.Vector, lambda vec: (cq.Vector, vec.toTuple()))
    copyreg.pickle(OCP.gp.gp_Trsf, _reduce_transform)
    copyreg.pickle(OCP.gp.gp_GTrsf, _reduce_Gtransform)
    copyreg.pickle(
        cq.Location, lambda loc: (cq.Location, (loc.wrapped.Transformation(),))
    )
    copyreg.pickle(OCP.TopoDS.TopoDS_Compound, _reduce_compound)
    copyreg.pickle(OCP.TopoDS.TopoDS_Solid, _reduce_solid)
    copyreg.pickle(OCP.TopoDS.TopoDS_Shell, _reduce_shell)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os
import shutil
import subprocess

from . import runtime_python
from . import logging as pc_logging


class CondaPythonRuntime(runtime_python.PythonRuntime):
    def __init__(self, ctx, version=None):
        super().__init__(ctx, "conda", version)

        self.conda_path = shutil.which("conda")

    def run(self, cmd, stdin=""):
        with self.lock:
            if not self.initialized:
                with pc_logging.Action("Conda", "create", self.version):
                    if self.conda_path is None:
                        raise Exception(
                            "ERROR: PartCAD is configured to use conda, but conda is missing"
                        )

                    try:
                        os.makedirs(self.path)

                        # Install new conda environment with the preferred Python version
                        p = subprocess.Popen(
                            [
                                self.conda_path,
                                "create",
                                "-y",
                                "-q",
                                "--json",
                                "-p",
                                self.path,
                                "python=%s" % self.version,
                            ],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                        _, stderr = p.communicate()
                        if not stderr is None and stderr != b"":
                            pc_logging.error("conda env install error: %s" % stderr)

                        # Install pip into the newly created conda environment
                        p = subprocess.Popen(
                            [
                                self.conda_path,
                                "install",
                                "-y",
                                "-q",
                                "--json",
                                "-p",
                                self.path,
                                "pip",
                            ],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                        _, stderr = p.communicate()
                        if not stderr is None and stderr != b"":
                            pc_logging.error("conda pip install error: %s" % stderr)

                        self.initialized = True
                    except Exception as e:
                        shutil.rmtree(self.path)
                        raise e

        return super().run(
            [
                self.conda_path,
                "run",
                "--no-capture-output",
                "-p",
                self.path,
                "python",
                # "python%s" % self.version,  # This doesn't work on Windows
            ]
            + cmd,
            stdin,
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#

import random
import string


class ShapeConfiguration:
    def __init__(self, config):
        self.config = config
        if "name" in config:
            self.name = config["name"]
        else:
            name = "part" + "".join(
                random.choices(string.ascii_uppercase + string.digits, k=8)
            )
            self.name = name
            self.config["name"] = name

    @staticmethod
    def normalize(name, config):
        # Handle the case of the part being declared in the config
        # but not defined (a one liner like "part_name:").
        # TODO(clairbee): Revisit whether it's a bug or a feature
        #                 that this code allows to load undeclared scripts
        if config is None:
            config = {}

        # Instead of passing the name as a parameter,
        # enrich the configuration object
        # TODO(clairbee): reconsider passing the name as a parameter
        config["name"] = name
        config["orig_name"] = name

        return config



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import os

from . import runtime_python


class NonePythonRuntime(runtime_python.PythonRuntime):
    def __init__(self, ctx, version=None):
        super().__init__(ctx, "none", version)

        if not self.initialized:
            os.makedirs(self.path)
            self.initialized = True

    def run(self, cmd, stdin=""):
        return super().run(
            [
                "python",
                # "python%s" % self.version, # This doesn't work on Windows
            ]
            + cmd,
            stdin,
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import os

from . import project as p


class ImportConfiguration:
    def __init__(self, config_obj={}):
        self.config_obj = config_obj
        self.name = self.config_obj.get("name")
        self.import_config_type = self.config_obj.get("type")


class ProjectFactory(ImportConfiguration):
    def __init__(self, ctx, parent, import_config_obj):
        super().__init__(import_config_obj)
        self.ctx = ctx
        self.parent = parent

        if parent is None:
            self.config_path = ctx.config_path
            self.config_dir = ctx.config_dir
        else:
            self.config_path = parent.config_path
            self.config_dir = parent.config_dir

        # TODO(clairbee): Initialize the config object if necessary

    def _create(self, config):
        # TODO(clairbee): Finalize the config object if necessary
        self.project = p.Project(self.ctx, self.name, self.path)
        # Make the project config inherit some properties of the import config
        self.project.config_obj["type"] = self.import_config_type

    def _save(self):
        self.ctx.projects[self.name] = self.project



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from .. import partcad as pc


class Scene(pc.Assembly):
    def __init__(self):
        super().__init__()



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-20
#
# Licensed under Apache License, Version 2.0.

from . import logging as pc_logging

# import svglib.svglib as svglib
# import reportlab.graphics.renderPM as renderPM


class PluginExportPngReportlab:
    def is_supported(self):
        return True

    def export(self, project, svg_path, width, height, filepath):
        import importlib

        # Load necessary symbols
        svglib = importlib.import_module("svglib.svglib")
        if svglib is None:
            pc_logging.error('Failed to load "svglib". Aborting.')
            return

        renderPM = importlib.import_module("reportlab.graphics.renderPM")
        if renderPM is None:
            pc_logging.error('Failed to load "renderPM". Aborting.')
            return

        # Render the raster image
        drawing = svglib.svg2rlg(svg_path)
        scale_width = float(width) / float(drawing.width)
        scale_height = float(height) / float(drawing.height)
        scale = min(scale_width, scale_height)
        drawing.scale(scale, scale)
        drawing.width *= scale
        drawing.height *= scale
        if not project is None:
            project.ctx.ensure_dirs_for_file(filepath)
        renderPM.drawToFile(
            drawing,
            filepath,
            fmt="PNG",
            configPIL={"transparent": True},
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

from .user_config import user_config

from . import runtime_python_none
from . import runtime_python_pypy
from . import runtime_python_conda


def create(ctx, version, python_runtime=None):
    if python_runtime is None:
        python_runtime = user_config.python_runtime
    if python_runtime == "none":
        return runtime_python_none.NonePythonRuntime(ctx, version)
    elif python_runtime == "pypy":
        return runtime_python_pypy.PyPyPythonRuntime(ctx, version)
    elif python_runtime == "conda":
        return runtime_python_conda.CondaPythonRuntime(ctx, version)
    else:
        raise Exception(
            "ERROR: invalid python runtime type (sandbox type) %s" % python_runtime
        )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import part as p


class PartFactory:
    # TODO(clairbee): Make the next line work for part_factory_file only
    path: typing.Optional[str] = None
    part: p.Part

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        part_config: object,
        extension: str = "",
    ):
        self.ctx = ctx
        self.project = source_project
        self.target_project = target_project
        self.part_config = part_config
        self.name = part_config["name"]
        self.orig_name = part_config["orig_name"]

    def _create(self, part_config: object):
        self.part = p.Part(part_config)
        self.part.project_name = (
            self.target_project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for part_factory_file only
        if self.path:
            self.part.path = self.path
        self.part.instantiate = lambda part_self: self.instantiate(part_self)

        self.target_project.parts[self.name] = self.part

        self.ctx.stats_parts += 1



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.

import hashlib
import os
import pathlib
import subprocess
import sys
import threading

from . import runtime
from . import logging as pc_logging


class PythonRuntime(runtime.Runtime):
    def __init__(self, ctx, sandbox, version=None):
        if version is None:
            version = "%d.%d" % (sys.version_info.major, sys.version_info.minor)
        super().__init__(ctx, "python-" + sandbox + "-" + version)
        self.version = version

        # Runtimes are meant to be executed from dedicated threads, outside of
        # the asyncio event loop. So a threading lock is appropriate here.
        self.lock = threading.RLock()

    def run(self, cmd, stdin=""):
        p = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate(
            input=stdin.encode(),
            # TODO(clairbee): add timeout
        )

        stdout = stdout.decode()
        stderr = stderr.decode()

        # TODO(clairbee): remove the below when a better troubleshooting mechanism is introduced
        # f = open("/tmp/log", "w")
        # f.write("Completed: %s\n" % cmd)
        # f.write(" stdin: %s\n" % stdin)
        # f.write(" stderr: %s\n" % stderr)
        # f.write(" stdout: %s\n" % stdout)
        # f.close()

        return stdout, stderr

    def ensure(self, python_package):
        guard_path = os.path.join(
            self.path, ".partcad.installed." + python_package
        )
        with self.lock:
            if not os.path.exists(guard_path):
                with pc_logging.Action("PipInst", self.version, python_package):
                    self.run(["-m", "pip", "install", python_package])
                pathlib.Path(guard_path).touch()

    def prepare_for_package(self, project):
        # Check if this project has python requirements
        requirements_path = os.path.join(project.path, "requirements.txt")
        if os.path.exists(requirements_path):
            # See if it was already prepared once
            project_hash = hashlib.sha256(project.path.encode()).hexdigest()
            flag_filename = ".partcad.project." + project_hash
            flag_path = os.path.join(self.path, flag_filename)
            with self.lock:
                if not os.path.exists(flag_path) or os.path.getmtime(
                    requirements_path
                ) > os.path.getmtime(flag_path):
                    # Install requirements and remember when we did that
                    with pc_logging.Action(
                        "PipReqs", self.version, project.name
                    ):
                        self.run(
                            ["-m", "pip", "install", "-r", requirements_path]
                        )
                    pathlib.Path(flag_path).touch()



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-22
#
# Licensed under Apache License, Version 2.0.
#


# Merge render configs (to override project's config with the part's config)
def render_cfg_merge(a: dict, b: dict, path=[]):
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                a[key] = render_cfg_merge(a[key], b[key], path + [str(key)])
            if isinstance(a[key], list) and isinstance(b[key], list):
                a[key].extend(b[key])
            elif a[key] != b[key]:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


DEFAULT_RENDER_WIDTH = 1024
DEFAULT_RENDER_HEIGHT = 1024

DEFAULT_RENDER_SVG_OPTS = {
    "width": DEFAULT_RENDER_WIDTH,
    "height": DEFAULT_RENDER_HEIGHT,
    "showAxes": False,
    "projectionDir": [1, -1, 0.75],
    "strokeWidth": 0.25,
    "strokeColor": [64, 192, 64],
    "hiddenColor": [32, 64, 32],
    "showHidden": False,
    # "marginLeft": 0,
    # "marginTop": 0,
}



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-01
#
# Licensed under Apache License, Version 2.0.
#

import base64
import os
import pickle
import sys

from . import part_factory_python as pfp
from . import wrapper
from . import logging as pc_logging

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import (
    register as register_cq_helper,
)  # import this one for `pickle` to use


class PartFactoryBuild123d(pfp.PartFactoryPython):
    def __init__(self, ctx, source_project, target_project, part_config):
        with pc_logging.Action(
            "InitBuild123d", target_project.name, part_config["name"]
        ):
            super().__init__(ctx, source_project, target_project, part_config)
            # Complement the config object here if necessary
            self._create(part_config)

    def instantiate(self, part):
        with pc_logging.Action("Build123d", part.project_name, part.name):
            wrapper_path = wrapper.get("build123d.py")

            request = {"build_parameters": {}}
            picklestring = pickle.dumps(request)
            request_serialized = base64.b64encode(picklestring).decode()

            self.runtime.ensure("build123d")
            response_serialized, errors = self.runtime.run(
                [wrapper_path, self.path], request_serialized
            )
            sys.stderr.write(errors)

            response = base64.b64decode(response_serialized)
            result = pickle.loads(response)

            if not result["success"]:
                pc_logging.error(result["exception"])
                raise Exception(result["exception"])

            self.ctx.stats_parts_instantiated += 1

            return result["shape"]



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-22
#
# Licensed under Apache License, Version 2.0.
#

import os

from . import part as p
from . import part_factory as pf
from . import logging as pc_logging


class PartFactoryFile(pf.PartFactory):
    part: p.Part

    def __init__(
        self, ctx, source_project, target_project, part_config, extension=""
    ):
        super().__init__(ctx, source_project, target_project, part_config)

        if "path" in part_config:
            self.path = part_config["path"]
        else:
            self.path = self.orig_name + extension

        if not os.path.isdir(source_project.config_dir):
            raise Exception(
                "ERROR: The project config directory must be a directory, found: '%s'"
                % source_project.config_dir
            )
        self.path = os.path.join(source_project.config_dir, self.path)
        if not os.path.isfile(self.path):
            raise Exception(
                "ERROR: The part path (%s) must be a file" % self.path
            )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-02-19
#
# Licensed under Apache License, Version 2.0.
#

from . import factory


class ShapeFactory(factory.Factory):
    def __init__(self) -> None:
        super.__init__()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-02-04
#
# Licensed under Apache License, Version 2.0.

import logging
from logging import DEBUG, INFO, WARN, WARNING, ERROR, CRITICAL
import threading
import time

setLevel = lambda *args, **kwargs: logging.getLogger("partcad").setLevel(
    *args, **kwargs
)
log = lambda *args, **kwargs: logging.getLogger("partcad").log(*args, **kwargs)
debug = lambda *args, **kwargs: logging.getLogger("partcad").debug(
    *args, **kwargs
)
info = lambda *args, **kwargs: logging.getLogger("partcad").info(
    *args, **kwargs
)
warn = lambda *args, **kwargs: logging.getLogger("partcad").warn(
    *args, **kwargs
)
warning = lambda *args, **kwargs: logging.getLogger("partcad").warning(
    *args, **kwargs
)
error = lambda *args, **kwargs: logging.getLogger("partcad").error(
    *args, **kwargs
)
critical = lambda *args, **kwargs: logging.getLogger("partcad").critical(
    *args, **kwargs
)
exception = lambda *args, **kwargs: logging.getLogger("partcad").exception(
    *args, **kwargs
)


def default_process_start(self_ops, op: str, package: str, item: str = None):
    if item is None:
        info("Starting process: %s: %s" % (op, package))
    else:
        info("Starting process: %s: %s: %s" % (op, package, item))


def default_process_end(self_ops, op: str, package: str, item: str = None):
    _ignore = True


def default_action_start(self_ops, op: str, package: str, item: str = None):
    if item is None:
        info("Starting action: %s: %s" % (op, package))
    else:
        info("Starting action: %s: %s: %s" % (op, package, item))


def default_action_end(self_ops, op: str, package: str, item: str = None):
    _ignore = True


# Dependency injection point for logging plugins
class Ops:
    process_start = default_process_start
    process_end = default_process_end
    action_start = default_action_start
    action_end = default_action_end


ops = Ops()


# Only one process at a time, no recursion
process_lock = threading.Lock()


# classes to be used with "with()" for blocks to be logged
class Process(object):
    def __init__(self, op: str, package: str, item: str = None):
        self.op = op
        self.package = package
        self.item = item
        self.succeeded = False
        self.start = 0.0

    async def __aenter__(self):
        self.__enter__()

    def __enter__(self):
        global process_lock

        if process_lock.acquire():
            self.start = time.time()
            ops.process_start(self.op, self.package, self.item)
            self.succeeded = True
        else:
            error("Nested process is detected. Status reporting is invalid.")
            self.succeeded = False

    async def __aexit__(self, *args):
        self.__exit__(*args)

    def __exit__(self, *_args):
        global process_lock
        global info

        if self.succeeded:
            process_lock.release()
            ops.process_end(self.op, self.package, self.item)

            delta = time.time() - self.start
            if self.item is None:
                info("DONE: %s: %s: %.2fs" % (self.op, self.package, delta))
            else:
                info(
                    "DONE: %s: %s: %s: %.2fs"
                    % (self.op, self.package, self.item, delta)
                )


class Action(object):
    def __init__(self, op: str, package: str, item: str = None):
        self.op = op
        self.package = package
        self.item = item

    async def __aenter__(self):
        self.__enter__()

    def __enter__(self):
        ops.action_start(self.op, self.package, self.item)

    async def __aexit__(self, *args):
        self.__exit__(*args)

    def __exit__(self, *_args):
        ops.action_end(self.op, self.package, self.item)
